@@ -94,6 +94,10 @@ static int verbose = 1;
 
 static void context_state_callback(pa_context *c, void *userdata);
 
+static void playback_stream_drain(struct userdata *u);
+
+static void playback_stream_drain_cb(pa_stream *s, int success, void *userdata);
+
 void pacat_log(const char *fmt, ...) {
     va_list args;
     va_start(args, fmt);
@@ -200,9 +204,6 @@ static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_l
     void *buffer = NULL;
     assert(s);
 
-    if (!max_length)
-        return;
-
     if ((space_in_vchan = libvchan_data_ready(u->play_ctrl)) < 0) {
         pacat_log("libvchan_data_ready() failed: return value %d", space_in_vchan);
         quit(u, 1);
@@ -211,7 +212,7 @@ static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_l
 
     buffer_length = (size_t)space_in_vchan > max_length ? max_length : (size_t)space_in_vchan;
     if (!buffer_length)
-        return;
+        goto maybe_cork;
 
     if (pa_stream_begin_write(s, &buffer, &buffer_length) || !buffer) {
         pacat_log("pa_stream_begin_write() failed: %s", pa_strerror(pa_context_errno(u->context)));
@@ -241,12 +242,88 @@ static void process_playback_data(struct userdata *u, pa_stream *s, size_t max_l
             quit(u, 1);
             return;
         }
+        space_in_vchan -= (int)buffer_length;
     } else
         pa_stream_cancel_write(s);
 
+maybe_cork:
+    if (u->pending_play_cork && space_in_vchan == 0) {
+        pacat_log("Cork requested and playback vchan empty. Draining playback stream.");
+        playback_stream_drain(u);
+    }
+
     return;
 }
 
+static void playback_stream_drain(struct userdata *u) {
+    pa_operation *o;
+    pa_stream *s = u->play_stream;
+
+    if (u->draining)
+        return; /* Cannot have more than one drain operation ongoing */
+    u->draining = true;
+
+    if (!(o = pa_stream_drain(s, playback_stream_drain_cb, u))) {
+        pacat_log("pa_stream_drain(): %s", pa_strerror(pa_context_errno(pa_stream_get_context(s))));
+        quit(u, 1);
+        return;
+    }
+
+    pa_operation_unref(o);
+}
+
+static void playback_stream_drain_cb(pa_stream *s, int success, void *userdata) {
+    struct userdata *u = userdata;
+    assert(s == u->play_stream);
+    u->draining = false;
+    /* Check that the draining was successful. */
+    if (!success)
+        return;
+
+    /*
+     * Check that the agent has not sent QUBES_PA_SINK_UNCORK_CMD in the
+     * meantime. This ensures correct behavior in the following situation:
+     *
+     * 1. Agent sends QUBES_PA_SINK_CORK_CMD.
+     * 2. Vchan is drained.
+     * 3. pa_stream_drain() is called.
+     * 4. Agent sends QUBES_PA_SINK_UNCORK_CMD.
+     * 5. PulseAudio calls playback_stream_drain_cb().
+     *
+     * The stream must not get corked in this case.
+     */
+    if (!u->pending_play_cork)
+        return;
+
+    /*
+     * Check that the vchan is still empty before actually corking the stream.
+     * This ensures correct behavior in the following situation:
+     *
+     * 1. Agent sends QUBES_PA_SINK_CORK_CMD.
+     * 2. Vchan is drained.
+     * 3. pa_stream_drain() is called.
+     * 4. Agent sends QUBES_PA_SINK_UNCORK_CMD.
+     * 5. Agent sends more data on the vchan.
+     * 6. Agent sends QUBES_PA_SINK_CORK_CMD.
+     * 7. PulseAudio calls playback_stream_drain_cb().
+     *
+     * Without this check, the data sent in step 5 would be (wrongly) left in
+     * the vchan.  If the agent does not send any more data, pa_stream_drain()
+     * will be called again when the vchan does become empty, so the stream will
+     * eventually be corked.
+     */
+    int space_in_vchan = libvchan_data_ready(u->play_ctrl);
+    if (space_in_vchan == 0) {
+        pacat_log("Playback vchan empty and playback stream drained. Corking playback stream.");
+        u->pending_play_cork = false;
+        pa_stream_cork(u->play_stream, 1, NULL, u);
+    } else if (space_in_vchan < 0) {
+        pacat_log("libvchan_data_ready() failed: return value %d", space_in_vchan);
+        quit(u, 1);
+        return;
+    }
+}
+
 /* This is called whenever new data may be written to the stream */
 static void stream_write_callback(pa_stream *s, size_t length, void *userdata) {
     struct userdata *u = userdata;
@@ -383,13 +460,25 @@ static void vchan_rec_callback(pa_mainloop_api *UNUSED(a),
                 g_mutex_unlock(&u->prop_mutex);
                 break;
             case QUBES_PA_SINK_CORK_CMD:
-                pacat_log("Stream cork");
-                pa_stream_cork(u->play_stream, 1, NULL, u);
+                u->pending_play_cork = true;
+                if (libvchan_data_ready(u->play_ctrl) > 0) {
+                    pacat_log("Deferred stream drain");
+                } else {
+                    pacat_log("Stream drain");
+                    playback_stream_drain(u);
+                }
                 break;
             case QUBES_PA_SINK_UNCORK_CMD:
                 pacat_log("Stream uncork");
+                u->pending_play_cork = false;
+                /*
+                 * do not clear u->draining, as draining while a drain
+                 * operation is in progress is not safe
+                 */
                 pa_stream_cork(u->play_stream, 0, NULL, u);
                 break;
+            default:
+                break;
         }
     }
     /* send the data if space is available */
@@ -919,18 +1008,25 @@ static void vchan_rec_async_connect(pa_mainloop_api *UNUSED(a),
     u->mainloop_api->io_free(u->rec_ctrl_event);
     u->rec_ctrl_event = NULL;
 
-    u->never_block = libvchan_buffer_space(u->rec_ctrl) > 8192;
+    if (u->never_block == -1)
+        u->never_block = libvchan_buffer_space(u->rec_ctrl) > 8192;
 
     /* when both vchans are connected, connect to the daemon */
     if (libvchan_is_open(u->play_ctrl) == VCHAN_CONNECTED)
         connect_pa_daemon(u);
 }
 
-static _Noreturn void usage(char *arg0) {
-    fprintf(stderr, "usage: %s [-l] [--] domid domname\n",
+static _Noreturn void usage(char *arg0, int arg) {
+    FILE *stream = arg ? stderr : stdout;
+    fprintf(stream, "usage: %s [-l] [--] domid domname\n",
             arg0 ? arg0 : "pacat-simple-vchan");
-    fprintf(stderr, "  -l - low-latency mode (higher CPU usage)\n");
-    exit(1);
+    fprintf(stream, "  -l - low-latency mode (higher CPU usage)\n");
+    fprintf(stream, "  -n - never block on vchan I/O (default if record buffer size is 8192 bytes or more)\n");
+    fprintf(stream, "  -b - always block on vchan I/O (default if record buffer size is less than 8192 bytes)\n");
+    fprintf(stream, "  -h - print this message\n");
+    if (fflush(NULL) || ferror(stdout) || ferror(stderr))
+        exit(1);
+    exit(arg);
 }
 
 int main(int argc, char *argv[])
@@ -943,19 +1039,32 @@ int main(int argc, char *argv[])
     char *pidfile_path;
     int pidfile_fd;
     int play_watch_fd, rec_watch_fd;
-    int i = 1;
+    int i;
 
-    if (argc <= 2 || argc > 5)
-        usage(argv[0]);
-    if (!strcmp(argv[i], "-l")) {
-        bufattr = &custom_bufattr;
-        i++;
+    memset(&u, 0, sizeof(u));
+    u.never_block = -1;
+    if (argc <= 2)
+        usage(argv[0], 1);
+    while ((i = getopt(argc, argv, "+lnbh")) != -1) {
+        switch (i) {
+            case 'l':
+                bufattr = &custom_bufattr;
+                break;
+            case 'n':
+                u.never_block = 1;
+                break;
+            case 'b':
+                u.never_block = 0;
+                break;
+            case 'h':
+                usage(argv[0], 0);
+            default:
+                usage(argv[0], 1);
+        }
     }
-    if (!strcmp(argv[i], "--"))
-        i++;
-    if (argc - i != 2)
-        usage(argv[0]);
-    const char *domid_str = argv[i], *domname = argv[i + 1];
+    if (argc - optind != 2)
+        usage(argv[0], 1);
+    const char *domid_str = argv[optind], *domname = argv[optind + 1];
     char *endptr;
     errno = 0;
     long l_domid = strtol(domid_str, &endptr, 10);
@@ -972,7 +1081,6 @@ int main(int argc, char *argv[])
         /* error already printed by create_pidfile() */
         exit(1);
 
-    memset(&u, 0, sizeof(u));
     u.ret = 1;
 
     g_mutex_init(&u.prop_mutex);
