@@ -29,14 +29,16 @@ struct userdata {
     pa_io_event* play_ctrl_event;
     pa_io_event* rec_ctrl_event;
 
-    int rec_allowed;
-    int rec_requested;
     GMutex prop_mutex;
     qdb_handle_t qdb;
     char *qdb_path;
     int control_socket_fd;
     pa_io_event* control_socket_event;
-    bool never_block;
+    bool rec_allowed;
+    bool rec_requested;
+    int8_t never_block;
+    bool pending_play_cork;
+    bool draining;
 };
 
 void pacat_log(const char *fmt, ...);
